#
# Utilities
#

enquote <- function(x) as.call(list(as.name("quote"), x))
docall <- function(fun, args) do.call("fun", lapply(args, enquote))


#
# Slave Loop Function
#

slaveLoop <- function(master) {
    repeat {
        tag <- recvTag(master)
	cat(paste("Tag:", tag, "\n"))
        data <- recvData(master)

        if (tag == "DONE") {
            closeNode(master)
            break;
        }
        else if (tag == "EXEC")
            lastValue <- try(eval(data))
        else if (tag == "LASTVAL")
            sendData(master, try(lastValue))
    }
}


#
# Socket Implementation
#

newSOCKnode <- function(machine = "localhost", ...,
                        options = defaultClusterOptions) {
    # **** allow some form of spec here
    # **** make sure options are quoted
    options <- addClusterOptions(options, ...)
    port <- getClusterOption("port", options)
    rhome <- getClusterOption("rhome", options)
    script <- file.path(rhome, "library", "snow", "RSOCKnode.sh")
    rshcmd <- getClusterOption("rshcmd", options)
    user <- getClusterOption("user", options)
    env <- paste("RPROG=", file.path(rhome, "bin", "R"),
                 " MASTER=", getClusterOption("master", options),
                 " PORT=", port,
                 " OUT=", getClusterOption("outfile", options),
                 " R_LIBS=", getClusterOption("rlibs", options),
                 sep="")
    system(paste(rshcmd, "-l", user, machine, "env", env, script))
    con <- socketConnection(port = port, server=TRUE, blocking=TRUE,
                            open="a+b")
    structure(list(con = con, host = machine), class = "SOCKnode")
}

makeSOCKmaster <- function() {
    master <- Sys.getenv("MASTER")
    port <- as.integer(Sys.getenv("PORT"))
    # maybe use `try' and slep/retry if first time fails?
    con <- socketConnection(master, port = port, blocking=TRUE, open="a+b")
    structure(list(con = con), class = "SOCKnode")
}

closeNode.SOCKnode <- function(node) close(node$con)

sendTag.SOCKnode <- function(node, tag) write(tag, node$con)
recvTag.SOCKnode <- function(node) readLines(node$con, 1)
sendData.SOCKnode <- function(node, data) serialize(data, node$con)
recvData.SOCKnode <- function(node) unserialize(node$con)


#
# PVM Implementation
#

newPVMnode <- function(where = "",
                       options = defaultClusterOptions) {
    # **** allow some form of spec here
    # **** make sure options are quoted
    rhome <- getClusterOption("rhome", options)
    script <- file.path(rhome, "library", "snow", "RPVMnode.sh")
    args <- c(paste("RPROG=", file.path(rhome, "bin", "R"), sep=""),
              paste("OUT=", getClusterOption("outfile", options), sep=""),
              paste("R_LIBS=", getClusterOption("rlibs", options), sep=""),
              script)
    tid <- .PVM.spawn(task="/usr/bin/env", arglist = args, where = where)
    structure(list(tid = tid, RECVTAG = 33,SENDTAG = 22), class = "PVMnode")
}

makePVMmaster <- function()
    structure(list(tid = .PVM.parent (), RECVTAG = 22, SENDTAG = 33),
              class = "PVMnode")

closeNode.PVMnode <- function(node) {}

sendTag.PVMnode <- function(node, tag) {
    .PVM.initsend ()
    .PVM.pkstrvec (tag)
    .PVM.send (node$tid, node$SENDTAG)
}

recvTag.PVMnode <- function(node) {
    .PVM.recv (node$tid, node$RECVTAG)
    .PVM.upkstrvec ()
}

# old versions using deparse/parse
sendData.PVMnode <- function(node, data) {
    ddata <- deparse(if (is.language(data)) enquote(data) else data)
    .PVM.initsend ()
    .PVM.pkstrvec (ddata)
    .PVM.send (node$tid, node$SENDTAG)
}

recvData.PVMnode <- function(node) {
    .PVM.recv (node$tid, node$RECVTAG)
    eval(parse(text = .PVM.upkstrvec ()))
}

# these require rpvm with serialization support
sendData.PVMnode <- function(node, data) {
    .PVM.initsend ()
    .PVM.serialize(data, node$con)
    .PVM.send (node$tid, node$SENDTAG)
}

recvData.PVMnode <- function(node) {
    .PVM.recv (node$tid, node$RECVTAG)
    .PVM.unserialize(node$con)
}


#
# Higher-Level Node Functions
#

closeNode <- function(node) UseMethod("closeNode")
sendTag <- function(node, tag) UseMethod("sendTag")
recvTag <- function(node) UseMethod("recvTag")
sendData <- function(node, data) UseMethod("sendData")
recvData <- function(node) UseMethod("recvData")

sendNode <- function(con, tag, value = NULL) {
    sendTag(con, tag)
    sendData(con, value)
    recvData(con)
}


postNode <- function(con, tag, value = NULL) {
    sendTag(con, tag)
    sendData(con, value)
}

stopNode <- function(n) {
    postNode(n, "DONE")
    closeNode(n)
}


#
#  Cluster Creation and Destruction
#

defaultClusterOptions <- NULL

#**** check valid cluster option

setDefaultClusterOptions <-
    function(port = 600011,
             master =  Sys.info()["nodename"],
             outfile = "/dev/null",
             rhome = Sys.getenv("R_HOME"),
             user = Sys.info()["user"],
             rshcmd = "ssh",
             rlibs = Sys.getenv("R_LIBS")) {
    defaultClusterOptions <<- new.env(parent=NULL)
    for (n in ls()) assign(n, get(n), env = defaultClusterOptions)
}

addClusterOptions <- function(options, ...) {
    new <- list(...)
    if (! is.null(new)) {
        options <- new.env(parent = options)
        names <- names(new)
        for (i in seq(along = new))
            assign(names[i], new[[i]], env = options)
    }
    options
}

getClusterOption <- function(name, options = defaultClusterOptions)
    get(name, env = options)

makePVMcluster <- function(count, ..., options = defaultClusterOptions) {
    if (! require(rpvm))
        stop("the `rpvm' package is needed for PVM clusters.")
    options <- addClusterOptions(options, ...)
    cl <- vector("list",count)
    for (i in seq(along=cl))
        cl[[i]] <- newPVMnode(options = options)
    cl
}

makeSOCKcluster <- function(names, ..., options = defaultClusterOptions) {
    if (! exists("serialize") && ! require(serialize))
        stop("the `serialize' package is needed for SOCK clusters.")
    options <- addClusterOptions(options, ...)
    cl <- vector("list",length(names))
    for (i in seq(along=cl))
        cl[[i]] <- newSOCKnode(names[[i]], options = options)
    cl
}

stopCluster <- function(cl)
    for (n in cl) stopNode(n)


#
# Cluster Functions
#

sendCall <- function(con, fun, args) {
    call <- as.call(list(as.name("docall"), fun, args))
    postNode(con, "EXEC", call)
}

clusterCall  <- function(cl, fun, ...) {
    for (i in seq(along = cl))
        sendCall(cl[[i]], fun, list(...))
    lapply(cl, sendNode, "LASTVAL")
}

clusterApply <- function(cl, x, fun, ...) {
    if (length(cl) < length(x))
        stop("data lengh must be at most cluster size")
    for (i in seq(along = x))
        sendCall(cl[[i]], fun, c(list(x[[i]]), list(...)))
    lapply(cl[seq(along=x)], sendNode, "LASTVAL")
}


clusterEvalQ<-function(cl, expr)
    clusterCall(cl, eval, substitute(expr), env=.GlobalEnv)

#
# Cluster SPRNG Support 
#
# adapted from rpvm (Li & Rossini)

clusterSetupSPRNG <- function (cl, seed = round(2^32 * runif(1)),
                            prngkind = "default", para = 0) 
{
    if (!is.character(prngkind) || length(prngkind) > 1)
        stop("'rngkind' must be a character string of length 1.")
    if (!is.na(pmatch(prngkind, "default")))
        prngkind <- "LFG"
    prngnames <- c("LFG", "LCG", "LCG64", "CMRG", "MLFG", "PMLCG")
    kind <- pmatch(prngkind, prngnames) - 1
    if (is.na(kind))
        stop(paste("'", prngkind, "' is not a valid choice", sep = ""))
    nc <- length(cl)
    invisible(clusterApply(cl, 0:(nc-1), initSprngNode, nc, seed, kind, para))
}

initSprngNode <- function (streamno, nstream, seed, kind, para) 
{
    if (! require(rsprng))
        stop("the `rsprng' package is needed for SPRNG support.")
    .Call("r_init_sprng", as.integer(kind), as.integer(streamno), 
        as.integer(nstream), as.integer(seed), as.integer(para))
    RNGkind("user")
}


#
# Parallel Functions
#

splitIndices <- function(nx, ncl) {
    batchsize <- if (nx %% ncl == 0) nx %/% ncl else 1 + nx %/% ncl
    batches <- (nx + batchsize - 1) %/% batchsize
    split(1:nx, rep(1:batches, each = batchsize)[1:nx])
}

splitIndices <- function(nx, ncl) {
    i <- 1:nx;
    structure(split(i, cut(i, ncl)), names=NULL)
}

splitList <- function(x, ncl)
    lapply(splitIndices(length(x), ncl), function(i) x[i])

splitRows <- function(x, ncl)
    lapply(splitIndices(nrow(x), ncl), function(i) x[i,, drop=F])

splitCols <- function(x, ncl)
    lapply(splitIndices(ncol(x), ncl), function(i) x[,i, drop=F])

parLapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitList(x, length(cl)), lapply, fun, ...))

parRapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitRows(x,length(cl)), apply, 1, fun, ...))

parCapply <- function(cl, x, fun, ...)
    docall(c, clusterApply(cl, splitCols(x,length(cl)), apply, 2, fun, ...))

parMM <- function(cl, A, B)
    docall(rbind,clusterApply(cl, splitRows(A, length(cl)), get("%*%"), B))

parSapply <- function (cl, X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE) 
{
    FUN <- match.fun(FUN)
    answer <- parLapply(cl,as.list(X), FUN, ...)
    if (USE.NAMES && is.character(X) && is.null(names(answer))) 
        names(answer) <- X
    if (simplify && length(answer) != 0) {
        common.len <- unique(unlist(lapply(answer, length)))
        if (common.len == 1) 
            unlist(answer, recursive = FALSE)
        else if (common.len > 1) 
            array(unlist(answer, recursive = FALSE),
                  dim = c(common.len, length(X)),
                  dimnames = list(names(answer[[1]]), names(answer)))
        else answer
    }
    else answer
}


#
#  Library Initialization
#

.First.lib <- function(libname, pkgname) {
    if (is.null(defaultClusterOptions))
	setDefaultClusterOptions()
}
